open Softcore 

type 'a u = {
  conj :    'a;
  disj :    'a;
  xor :     'a;
  equ :     'a;
  imp :     'a;
  closed :  'a;
  it_conj : 'a;
  it_disj : 'a;
  loop :    'a;
  lemmas :  'a;
  initial_shift : 'a;
  min_shift : 'a;
}

type t = int ref u

let init ~f ~upper_bound v = {
  conj          = f v;
  disj          = f v;
  xor           = f v;
  equ           = f v;
  imp           = f v;
  closed        = f v;
  it_conj       = f v;
  it_disj       = f v;
  loop          = f v;
  lemmas        = f v;
  initial_shift = ref ((function None -> 0 | Some x -> Indexes.get_shift x) upper_bound);
  min_shift     = f v;
}

let weird_map f = {
  conj          = (fun s -> f s.conj);
  disj          = (fun s -> f s.disj);
  xor           = (fun s -> f s.xor);
  equ           = (fun s -> f s.equ);
  imp           = (fun s -> f s.imp);
  closed        = (fun s -> f s.closed);
  it_conj       = (fun s -> f s.it_conj);
  it_disj       = (fun s -> f s.it_disj);
  loop          = (fun s -> f s.loop);
  lemmas        = (fun s -> f s.lemmas);
  initial_shift = (fun s -> f s.initial_shift);
  min_shift     = (fun s -> f s.min_shift);
}

let incr = weird_map incr
let set = weird_map (fun x v -> x := v)
let new_counter = init ~f:ref 0
let count = weird_map (fun x -> !x)
let update_min_shift t s = s.min_shift := min !(s.min_shift) (Indexes.get_shift t - 1)

let (^^) f g x = f x; g x

let sum l v = List.fold_left ~init:0 ~f:(fun acc f -> acc + f v) l

let prop_count = sum [count.conj; count.disj; count.xor; count.equ; count.imp;]
let it_count = sum [count.it_conj; count.it_disj;]

let to_string s = Printf.sprintf "
Applications of tableau rules:
/\\:  %d
\\/:  %d
(+): %d
<->: %d
->:  %d
Iterated /\\: %d
Iterated \\/:  %d
-------
Total propositional rules: %d
Total iterated rules:      %d

Number of closed leaves:  %d
Number of looping leaves: %d
Number of lemmas:         %d

Maximum number of unfoldings: %d 
(if you are surprised by this number, notice that the tableau is constructed depth-first)

"
  (count.conj s)
  (count.disj s)
  (count.xor s)
  (count.equ s)
  (count.imp s)
  (count.it_conj s)
  (count.it_disj s)
  (prop_count s)
  (it_count s)
  (count.closed s)
  (count.loop s)
  (count.lemmas s)
  (!(s.initial_shift) - !(s.min_shift))
